from pydantic import BaseModel


class AbstractModelCreateSchema(BaseModel):
    pass


class AbstractModelCreateErrorsSchema(BaseModel):
    pass


class AbstractModelGetSchema(BaseModel):
    pass


class AbstractModelUpdateSchema(BaseModel):
    pass


class AbstractModelUpdateErrorsSchema(BaseModel):
    pass



from sqlalchemy.ext.asyncio import AsyncSession, create_async_engine
from sqlalchemy.orm import sessionmaker

from architectonics.config.base_database_settings import base_database_settings

engine = create_async_engine(
    base_database_settings.DATABASE_CONNECTION_STRING,
    future=True,
    echo=base_database_settings.DEBUG,
)

session = sessionmaker(
    bind=engine,
    class_=AsyncSession,
    expire_on_commit=False,
)



class RepositoryException(Exception):
    """Base class for all repository exceptions."""


class ObjectNotFoundRepositoryException(RepositoryException):
    """Exception raised when can't find a row."""


class ForeignKeyViolationRepositoryException(RepositoryException):
    """Exception raised when foreign key constraint fails."""


class ObjectAlreadyExistsRepositoryException(RepositoryException):
    """Exception raised when unique constraint failes."""


class IntegrityErrorRepositoryException(RepositoryException):
    """Exception raised when integrity constraint fails."""



from starlette.status import HTTP_200_OK, HTTP_204_NO_CONTENT, HTTP_404_NOT_FOUND, HTTP_409_CONFLICT, HTTP_422_UNPROCESSABLE_ENTITY

from architectonics.models.abstract_base_model import AbstractBaseModel
from architectonics.repositories.exceptions import (
    ForeignKeyViolationRepositoryException,
    IntegrityErrorRepositoryException,
    ObjectAlreadyExistsRepositoryException,
    ObjectNotFoundRepositoryException,
)
from architectonics.repositories.repository import AbstractBaseRepository
from architectonics.services.schemas import (
    AbstractModelCreateSchema,
    AbstractModelUpdateSchema,
)
from architectonics.services.types import ErrorsType, StatusCodeType


class AbstractBaseService:
    _repository: AbstractBaseRepository = NotImplemented

    async def validate(self, **kwargs) -> dict[str, any]:

        if not kwargs:
            return {}

        attrs = {field: value for field, value in kwargs.items() if field in self._repository.model_fields}

        return attrs

    async def _validate_values(self, **kwargs) -> tuple[dict[str, any], dict[str, list[str]]]:

        errors = {}

        for key, value in kwargs.items():
            validation_method = f"validate_{key}"

            if hasattr(self, validation_method):
                result, error = getattr(
                    self,
                    validation_method,
                )(
                    value,
                    kwargs,
                )

                if not result:
                    errors[key] = errors.get(key, []) + error

        attrs = await self.validate(
            **kwargs,
        )

        return attrs, errors

    async def create_model(
        self,
        create_schema: AbstractModelCreateSchema,
    ) -> tuple[AbstractBaseModel | None, ErrorsType, StatusCodeType]:

        schema_dict = create_schema.model_dump()

        attrs, errors = await self._validate_values(
            **schema_dict,
        )

        if errors:
            return None, errors, HTTP_422_UNPROCESSABLE_ENTITY

        try:
            model = await self._repository.create_model(
                values=attrs,
            )
        except ObjectAlreadyExistsRepositoryException:
            return None, "object_already_exist", HTTP_409_CONFLICT

        return model, None, HTTP_200_OK

    async def get_model(
        self,
        model_id: str,
    ) -> tuple[AbstractBaseModel | None, ErrorsType, StatusCodeType]:

        try:
            model = await self._repository.get_model(
                model_id=model_id,
            )
        except ObjectNotFoundRepositoryException:
            return None, "object_not_found", HTTP_404_NOT_FOUND

        return model, None, HTTP_200_OK

    async def update_model(
        self,
        model_id: str,
        update_schema: AbstractModelUpdateSchema,
    ) -> tuple[AbstractBaseModel | None, ErrorsType, StatusCodeType]:

        schema_dict = update_schema.model_dump(
            by_alias=False,
        )

        attrs, errors = await self._validate_values(
            **schema_dict,
        )

        if errors:
            return None, errors, HTTP_422_UNPROCESSABLE_ENTITY

        try:
            model = await self._repository.update_model(
                model_id=model_id,
                values=attrs,
            )
        except ForeignKeyViolationRepositoryException as e:
            return None, f"{e}", HTTP_422_UNPROCESSABLE_ENTITY
        except IntegrityErrorRepositoryException as e:
            return None, f"{e}", HTTP_404_NOT_FOUND
        except ObjectNotFoundRepositoryException:
            return None, "object_not_found", HTTP_404_NOT_FOUND

        return model, None, HTTP_200_OK

    async def delete_model(
        self,
        model_id: str,
    ) -> tuple[None, ErrorsType, StatusCodeType]:

        _, errors, status_code = await self.get_model(
            model_id=model_id,
        )

        if errors:
            return None, errors, status_code

        await self._repository.delete_model(
            model_id=model_id,
        )

        return None, None, HTTP_204_NO_CONTENT

    async def get_models_list(
        self,
    ) -> tuple[list[AbstractBaseModel], ErrorsType, StatusCodeType]:

        models = await self._repository.get_models_list()

        return models, None, HTTP_200_OK



from abc import ABC
from typing import Callable

from asyncpg.exceptions import ForeignKeyViolationError
from sqlalchemy import delete, select, update
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from architectonics.config.database import session
from architectonics.models.abstract_base_model import AbstractBaseModel
from architectonics.repositories.exceptions import (
    ForeignKeyViolationRepositoryException,
    IntegrityErrorRepositoryException,
    ObjectAlreadyExistsRepositoryException,
    ObjectNotFoundRepositoryException,
)


class AbstractBaseRepository(ABC):
    _model: type[AbstractBaseModel] = NotImplemented
    _session: Callable = session
    _integrity_error: type[IntegrityError] = IntegrityError

    def get_session(self) -> AsyncSession:
        return self._session()

    @property
    def model_fields(self):
        return self._model.__table__.columns

    async def create_model(
        self,
        values: dict[str, any],
    ) -> AbstractBaseModel:

        model = self._model(
            **values,
        )

        async with self.get_session() as session:

            session.add(model)

            try:
                await session.commit()
            except self._integrity_error as e:
                raise ObjectAlreadyExistsRepositoryException(e)

            return model

    async def get_model(
        self,
        model_id: str,
    ) -> AbstractBaseModel:

        statement = select(
            self._model,
        ).where(
            self._model.id == model_id,
        )

        async with self.get_session() as session:

            result = await session.execute(
                statement=statement,
            )

            model = result.scalars().first()

            if model is None:
                raise ObjectNotFoundRepositoryException()

            return model

    async def update_model(
        self,
        model_id: str,
        values: dict[str, any],
    ) -> AbstractBaseModel:

        filtered_values = {k: v for k, v in values.items() if v is not None}

        statement = (
            update(
                self._model,
            )
            .where(
                self._model.id == model_id,
            )
            .values(
                **filtered_values,
            )
            .returning(
                self._model,
            )
        )

        async with self.get_session() as session:
            try:
                result = await session.execute(
                    statement=statement,
                )
                await session.commit()
            except self._integrity_error as e:

                orig = getattr(e.orig, "__cause__", None)

                if isinstance(orig, ForeignKeyViolationError):
                    raise ForeignKeyViolationRepositoryException()

                raise IntegrityErrorRepositoryException(e)

            model = result.scalars().first()

            if model is None:
                raise ObjectNotFoundRepositoryException()

            return model

    async def delete_model(
        self,
        model_id: str,
    ) -> None:

        statement = delete(
            self._model,
        ).where(
            self._model.id == model_id,
        )

        async with self.get_session() as session:
            await session.execute(statement)
            await session.commit()

    async def get_models_list(
        self,
    ) -> list[AbstractBaseModel]:

        statement = select(
            self._model,
        )

        async with self.get_session() as session:

            result = await session.execute(
                statement=statement,
            )

            models = result.scalars().all()

            return models



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['architectonics',
 'architectonics.config',
 'architectonics.models',
 'architectonics.repositories',
 'architectonics.schemas',
 'architectonics.services',
 'architectonics.views']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy==2.0.43',
 'aio-pika>=9.5.7,<10.0.0',
 'alembic==1.16.5',
 'asyncpg==0.30.0',
 'dotenv==0.9.9',
 'fastapi==0.116.1',
 'pydantic==2.11.7',
 'uvicorn==0.35.0']

setup_kwargs = {
    'name': 'architectonics',
    'version': '0.0.9',
    'description': '',
    'long_description': None,
    'author': 'kirshuvl',
    'author_email': 'kirshuvl@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
